import os
import re
from string import Template
import sys

as3_file = '../src/ld20/Embedded.as'
path_to_assets_directory_from_output = '../../assets'
path_to_assets_directory_from_workingdir = '.'
generated_section_start_re = re.compile(r'\s*/\* GENERATED CODE START \*/\s*', re.IGNORECASE)
generated_section_end_re = re.compile(r'\s*/\* GENERATED CODE END \*/\s*', re.IGNORECASE)
ignore_folders = [ 'music' ]

bitmapdata_template = Template("""
        [Embed(source = '$path')]
        private static const _$name:Class;
        public static const $name:BitmapData = new _$name().bitmapData;""")
sound_template = Template("""
        [Embed(source = '$path')]
        private static const _$name:Class;
        public static const $name:Sound = new _$name();""")
xml_template = Template("""
        [Embed(source='$path', mimeType='application/octet-stream')]
        private static const _$name:Class;
        public static const $name:XML = byteArrayToXml(new _$name);""")
array_template = Template("""
        public static const ${prefix}List:Array = [$names];""");

extension_to_template_map = { '.png':bitmapdata_template,
                              '.mp3':sound_template,
                              '.tmx':xml_template }
extension_to_array_template_map = { '.png':array_template,
                                    '.mp3':array_template }

class EmbedArrayFinder:
    r"""if I have things named foo0, foo1, etc. autogenerate an array
    of their data.

    self.data maps from extensions to maps from prefixes to lists of numbers"""

    pattern = re.compile(r'([A-Za-z_]+)([0-9]+)')

    def __init__(self):
        self.data = {}

    def check(self, basename, ext):
        r"""run the filename by EmbedArrayFinder in case it's his kind of
        thing"""
        match = EmbedArrayFinder.pattern.match(basename)
        if match:
            prefix, number = match.group(1), match.group(2)
            ext = ext.lower()
            if ext not in self.data:
                self.data[ext] = {prefix:[]}
            if prefix not in self.data[ext]:
                self.data[ext][prefix] = []
            self.data[ext][prefix].append(number)

    def get_code(self):
        code = ''
        for ext, prefixes in self.data.items():
            if ext in extension_to_array_template_map:
                template = extension_to_array_template_map[ext]
                for prefix, numbers in prefixes.items():
                    if len(numbers) > 1:
                        numbers.sort()
                        nameslist = ','.join([prefix+x for x in numbers])
                        code += template.substitute( prefix=prefix,
                                                     names=nameslist )
        return code

    def clear(self):
        self.data = {}

def is_same_file_or_dir(a,b):
    return os.path.abspath(a).lower() == os.path.abspath(b).lower()

def get_path_prefix(path):
    r"""names get prefixed if they are deeper in the folder hierarchy than assets"""
    path = path.replace(path_to_assets_directory_from_workingdir, '')
    if len(path):
        if path[0]=='\\' or path[0]=='/':
            path = path[1:]
        path = re.sub(r'[\\/]', '_', path)
        return path+'_'
    else:
        return ''

def get_embed_code():
    embedcode = ''
    eaf = EmbedArrayFinder()
    for root, dirs, files in os.walk(path_to_assets_directory_from_workingdir):
        if all([not is_same_file_or_dir(root, ignored) for ignored in ignore_folders]):
            for file in files:
                name, ext = os.path.splitext(file)
                ext = ext.lower()
                if ext in extension_to_template_map:
                    name = get_path_prefix(root) + name
                    name = re.sub(r'[\[\]\{\}\(\), \+=\-`~!@#\$%\^&\.\']', '_', name)
                    embedpath = os.path.join(root.replace(path_to_assets_directory_from_workingdir, path_to_assets_directory_from_output), file).replace('\\', '/')
                    embedcode += extension_to_template_map[ext].substitute(path=embedpath, name=name)
                    eaf.check(name, ext)
            embedcode += eaf.get_code();
            eaf.clear()
    return embedcode

def edit_file_in_place(generatedcode):
    import fileinput
    within_generated_section = False
    for line in fileinput.input(as3_file, inplace=True):
        if not within_generated_section:
            sys.stdout.write(line)
            if generated_section_start_re.match(line):
                within_generated_section = True
                for generatedline in generatedcode.splitlines():
                    sys.stdout.write(generatedline + '\n')
        if within_generated_section:
            if generated_section_end_re.match(line):
                within_generated_section = False
                sys.stdout.write(line)

edit_file_in_place(get_embed_code())



