package ld20.gamestate
{
    import flash.geom.Point;
    import ld20.gameobject.Player;
    import ld20.rooms.RoomData;
    import ld20.rooms.RoomLoader;
    import stego.collections.LinkedListNode;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.gameobject.TileData;
    import stego.gameobject.TileMap;
	import stego.GameState;
    import stego.script.Step;
    import stego.script.step.Loop;
    import stego.ui.Fader;
    import stego.util.collision.CollisionInfo;
    import stego.util.MathUtil;
    import stego.util.scene.GameObjectHolder;

    public class Playing extends GameState
    {
        protected var roomLoader:RoomLoader;
        protected var uiLayer:GameObjectHolder;
        protected var levelBeaten:Boolean;
        protected var levelTransitionInProgress:Boolean;
        protected static const levelFadeOutTime:Number = 0.25;
        protected static const levelFadeInTime:Number = 0.25;
        protected var fader:Fader;

        protected static const maxPlayers:Number = 8;
        protected static const spawnPeriod:Number = 0.25;
        protected static var spawnCountdown:Number = 0;

        override public function begin(game:GameInterface):void
        {
            super.begin(game);

            game.zoom = 3;
            game.cameraX = 0.5 * game.stageWidth / 3.0;
            game.cameraY = 0.5 * game.stageHeight / 3.0;

            levelBeaten = false;
            levelTransitionInProgress = false;

            roomLoader = new RoomLoader();
            roomLoader.init(game);
            roomLoader.loadRoom(0);

            game.physics.addCollisionPair(roomLoader.playerLayer,roomLoader.playerLayer);
            game.physics.addCollisionPair(roomLoader.playerLayer,roomLoader.itemLayer);
            game.physics.addCollisionPair(roomLoader.playerLayer,roomLoader.enemyLayer);

            uiLayer = game.layers.addLayer();

            fader = uiLayer.add(Fader) as Fader;
            fader.fadeToColor(0, 0xe6e62e);
            fader.fadeBack(0.25);

            game.cheats.addCheatCode("+", startLevelChange);
        }

        override public function update():void
        {
            super.update();

            // spawn the players
            playerSpawning();

            if (!levelBeaten)
            {
                if (0 == roomLoader.enemyLayer.objects.count)
                {
                    doorOpening();
                    levelBeaten = true;
                }
            }

            killOffScreenBaddies();
        }

        protected function killOffScreenBaddies():void
        {
            var node:LinkedListNode = roomLoader.enemyLayer.objects.first;
            while (node)
            {
                var go:GameObject = node.value;
                go.alive = go.isOnScreen;
                node = node.next;
            }
        }

        protected function doorOpening():void
        {
            // loop over the tiles. see if any are doors
            var tm:TileMap = roomLoader.tileMap;

            for (var col:int = 0; col < tm.getNumColumns(); ++col )
            {
                for (var row:int = 0; row < tm.getNumRows(); ++row )
                {
                    // open if it is time for that
                    var td:TileData = tm.getTileData(col, row, false);
                    var userdata:Object = td.userData;
                    if (userdata.hasOwnProperty("exit"))
                    {
                        var exitIndex:int = userdata["exit"];
                        if (td.index != exitIndex)
                        {
                            tm.setTile(col, row, exitIndex, userdata);
                        }
                    }
                }
            }
        }

        override public function collision():void
        {
            super.collision();

            var node:LinkedListNode = roomLoader.playerLayer.objects.first;
            while (node)
            {
                roomLoader.tileMap.collide(node.value, game.time.deltaTime, playerTileMapCollision);
                node = node.next;
            }

            node = roomLoader.enemyLayer.objects.first;
            while (node)
            {
                roomLoader.tileMap.collide(node.value, game.time.deltaTime);
                node = node.next;
            }
        }

        /** react to the player hitting a tile */
        protected function playerTileMapCollision( tileData:TileData,
                                                   tileMap:TileMap,
                                                   go:GameObject,
                                                   ci:CollisionInfo ):void
        {
            if (levelBeaten && !levelTransitionInProgress)
            {
                var userdata:Object = tileData.userData;
                if (userdata.hasOwnProperty("exit"))
                {
                    startLevelChange();
                }
            }
        }

        protected function startLevelChange():void
        {
            fader.fadeToColor(levelFadeOutTime, 0, changeLevelAndFadeIn);
            levelTransitionInProgress = true;
        }
        protected function changeLevelAndFadeIn():void
        {
            roomLoader.loadNextRoom();
            levelTransitionInProgress = false;
            levelBeaten = false;
            fader.fadeBack(levelFadeOutTime);
        }

        /** make sure we always have players */
        protected function playerSpawning():void
        {
            switch (roomLoader.spawnType)
            {
                case RoomLoader.SPAWNTYPE_ONE_AT_A_TIME:
                    if (0 == roomLoader.playerLayer.objects.count)
                    {
                        spawnPlayer();
                    }
                    break;
                case RoomLoader.SPAWNTYPE_CONTINUOUS:
                    if ( ( maxPlayers > roomLoader.playerLayer.objects.count) &&
                         (!roomLoader.playerLayer.anyOverlappingRect(roomLoader.spawnArea)) )
                    {
                        if (spawnCountdown <= 0)
                        {
                            spawnPlayer();
                            spawnCountdown += spawnPeriod;
                        }

                        spawnCountdown -= game.time.deltaTime;
                    }
                    else
                    {
                        spawnCountdown = spawnPeriod;
                    }
                    break;
                default:
                    throw new Error("What is spawn type \""+roomLoader.spawnType+"\"");
                    break;
            }
        }

        /** add a player */
        protected function spawnPlayer():void
        {
            var player:Player = roomLoader.playerLayer.add(Player) as Player;
            player.x = roomLoader.spawnPoint.x;
            player.y = roomLoader.spawnPoint.y;
        }

        override public function end():void
        {
            super.end();
            game.physics.removeAllCollisionPairs();
            roomLoader.shutdown();
            uiLayer.removeAll();
            uiLayer = null;
            game.resetCamera();
            game.cheats.removeCheatCode("+", startLevelChange);
        }
    }
}

