package stego.behavior.platformer
{
    import stego.behavior.Base;
    import stego.constants.Keyboard;
    import stego.GameObject;
    import stego.util.InputUtil;

    /**
     * Behavior for jumping.  Made to go nice with Run.
     * @author Mike Meyer
     */
    public class Jump extends Base
    {
        /** array of keys than you can use to jump */
        public var jumpKeys:Array = [Keyboard.SPACE, Keyboard.SHIFT, Keyboard.UP, Keyboard.Z, Keyboard.X, Keyboard.W];

        /** magic number used to store the mouse in _jumpKey */
        protected static const MOUSE_BUTTON_KEYCODE:uint = 255;

        /** whether the mouse button should also be considered a jump key */
        public var mouseButtonJumps:Boolean = false;

        /** remember which key they used to jump, so we know which key's release event we care about */
        protected var _jumpKey:uint = 0;

        /** whether the character is currently in the middle of a jump */
        public var isJumping:Boolean = false;

        /**
         * how many times the character can jump when already in the air
         * can be negative one if you want infinite air jumps
         */
        public var numAirJumps:int = 0;

        /** how many air jumps have been used. gets reset when he hits the ground. */
        protected var _airJumpsUsed:int = numAirJumps;

        /** how fast to go upward when he jumps */
        public var jumpSpeed:Number = 297;

        /** how fast to go slow down the upward motion when you let go of the jump button */
        public var jumpStopAccel:Number = 0;

        /** if set, this gets called with no args when the player jumps */
        public var jumpCallback:Function;

        /**
         * constructor
         * @param	inputUtil
         */
        public function Jump(inputUtil:InputUtil)
        {
            super(inputUtil);
        }

        override public function init(character:GameObject):void
        {
            super.init(character);
            if (null != _character)
            {
                _character.affectedByGravity = true;
            }
        }

        /**
         * process one frame of controls
         */
        override public function update(deltaTime:Number):void
        {
            if ( inputUtil.isUp(_jumpKey) || character.isOnGround)
            {
                isJumping = false;
                _jumpKey = 0;
            }

            // air jumps always get reset when you touch the ground
            if (character.isOnGround) { resetAirJumps(); }

            // if we aren't holding the jump button anymore, cut
            // off the jump height
            if (!character.isOnGround && !isJumping && (character.dy < 0) )
            {
                character.dy += jumpStopAccel * deltaTime;

                // don't shoot us down, though
                if (character.dy > 0)
                {
                    character.dy = 0;
                }
            }

            // starting a new jump
            var justPressedJumpKey:uint = checkForJumpPress();
            if ( (justPressedJumpKey != 0) &&
                 jumpAllowed() )
            {
                doJump(justPressedJumpKey);
            }
        }

        /**
         * check for the press of a jump button
         * return it's keycode, MOUSE_BUTTON_KEYCODE, or zero if none pressed
         */
        protected function checkForJumpPress():uint
        {
            if (mouseButtonJumps && _inputUtil.mouseButtonJustPressed)
            {
                return MOUSE_BUTTON_KEYCODE;
            }
            for each (var keyCode:uint in jumpKeys)
            {
                if (_inputUtil.wasJustPressed(keyCode))
                {
                    return keyCode;
                }
            }
            return 0;
        }

        /**
         * overridable handling of the actual jumping
         */
        protected function doJump(justPressedJumpKey:uint):void
        {
            // consumes an airjump if not on ground
            if (!character.isOnGround) { ++_airJumpsUsed; }

            isJumping = true;
            _jumpKey = justPressedJumpKey;

            character.dy = -jumpSpeed;

            if (null != jumpCallback)
            {
                jumpCallback();
            }
        }

        /** overridable criteria for whether you are allowed to jump */
        protected function jumpAllowed():Boolean
        {
            return character.isOnGround || (_airJumpsUsed < numAirJumps);
        }

        /**
         * reset the number of air jumps left to the max
         */
        public function resetAirJumps():void
        {
            _airJumpsUsed = 0;
        }
    }
}
