﻿package stego.debug
{
    import flash.display.BitmapData;
    import flash.events.Event;
    import flash.geom.Matrix;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import flash.utils.getQualifiedClassName;
    import stego.collections.LinkedList;
    import stego.collections.LinkedListNode;
    import stego.constants.Colors;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.gameobject.CanvasGameObject;
    import stego.ui.TextButton;
    import stego.ui.TextField;
    import stego.util.collision.HitBox;
    import stego.util.graphics.Canvas;
    import stego.util.MathUtil;
    import stego.util.ObjectPool;
    import stego.util.scene.GameObjectHolder;
    import stego.util.StringUtil;

    /**
     * debugging tools
     * @author Mike Meyer
     */
    public class DebugWidget extends GameObject
    {
        protected var _log:GameObject;
        protected var _profileTab:GameObject;
        protected var _logTextField:TextField;
        protected var _poolStatus:TextField;
        protected var _reusablesStatus:TextField;
        protected var _profilingStatus:TextField;
        protected var _hitboxLayer:CanvasGameObject;

        protected static const BACKGROUND_ALPHA:Number = 0.6;

        /** pixels of extra space between the buttons */
        protected static const BUTTON_SPACING:Number = 20;

        /** pixels betwee the left side of the screen and where debug text starts */
        protected static const LEFT_MARGIN:Number = 10;

        /**
         *
         */
        override public function init(game:GameInterface):void
        {
            super.init(game);

            if (0 == numChildren) // hack to see if it has been init'ed
            {
                createBackground(game.stageWidth, game.stageHeight);
                createTabBar();
                _log = createLog();
                _poolStatus = createStatusTextField();
                _reusablesStatus = createStatusTextField();
                _profileTab = createProfileTab();
                _hitboxLayer = internalGameObjects.addAtIndex(1, CanvasGameObject) as CanvasGameObject;
                _hitboxLayer.width = game.stageWidth;
                _hitboxLayer.height = game.stageHeight;

                internalGameObjects.add(FpsCounter);
            }

            // default to log view
            goToLog();

            // starts out invisible
            this.visible = false;
        }

        /**
         * make the background to go behind everything
         */
        protected function createBackground( width:Number, height:Number ):void
        {
            graphics.beginFill(0x000000, BACKGROUND_ALPHA);
            graphics.drawRect(
                0,
                0,
                width,
                height );
            graphics.endFill();
        }

        /**
         * create the clickable list of tabs
         */
        protected function createTabBar():void
        {
            var button:GameObject;
            button = addButton(LEFT_MARGIN, "log", goToLog);
            button = addButton(button.x + button.width + BUTTON_SPACING, "pool", goToPoolStatus);
            button = addButton(button.x + button.width + BUTTON_SPACING, "reuse", goToReusablesStatus);
            button = addButton(button.x + button.width + BUTTON_SPACING, "profiling", goToProfiling);
            button = addButton(button.x + button.width + BUTTON_SPACING, "hitboxes", goToHitboxes);
        }

        /**
         * add a tab button
         */
        private function addButton(x:Number, label:String, callback:Function):GameObject
        {
            return addButtonTo(internalGameObjects, x, 20, label, callback);
        }

        /**
         * add a text to something
         */
        private function addButtonTo(holder:GameObjectHolder, x:Number, y:Number, label:String, callback:Function):GameObject
        {
            var tb:TextButton = holder.add(TextButton) as TextButton;
            tb.setText(label);
            tb.x = x;
            tb.y = y;
            tb.setFontColor(Colors.WHITE);
            tb.setHoverColor(Colors.YELLOW);
            tb.callback = callback;
            return tb;
        }

        /**
         * react to press of the log button
         */
        private function goToLog():void
        {
            hideAllTabs();
            _log.visible = true;
        }

        /**
         * react to press of the pool button
         */
        private function goToPoolStatus():void
        {
            hideAllTabs();
            _poolStatus.visible = true;
        }

        /**
         * react to press of the reusables button
         */
        private function goToReusablesStatus():void
        {
            hideAllTabs();
            _reusablesStatus.visible = true;
        }

        /**
         * react to press of the profiling button
         */
        private function goToProfiling():void
        {
            hideAllTabs();
            _profilingStatus.visible = true;
        }

        /**
         *
         */
        private function goToHitboxes():void
        {
            hideAllTabs();
            _hitboxLayer.visible = true;
        }

        /**
         * hide all the objects that are in these half-assed tabs
         */
        protected function hideAllTabs():void
        {
            _log.visible = false;
            _poolStatus.visible = false;
            _reusablesStatus.visible = false;
            _profilingStatus.visible = false;
            _hitboxLayer.visible = false;
        }

        /**
         * make the log display
         */
        protected function createLog():GameObject
        {
            if (!_log)
            {
                _log = internalGameObjects.add();
                _log.y = 60;
                addButtonTo(_log.internalGameObjects, LEFT_MARGIN, 0, "clear", clearLog);

                _logTextField = _log.internalGameObjects.add(TextField) as TextField;
                _logTextField.fontColor = Colors.WHITE;
                _logTextField.x = LEFT_MARGIN;
                _logTextField.y = 20;
            }

            return _log;
        }

        /**
         * build the profiling tab
         */
        private function createProfileTab():GameObject
        {
            if (!_profileTab)
            {
                _profileTab = internalGameObjects.add();
                _profileTab.y = 60;
                addButtonTo(_profileTab.internalGameObjects, LEFT_MARGIN, 0, "clear", Profiler.clearProfilingInfo);
                _profilingStatus = _profileTab.internalGameObjects.add(TextField) as TextField;
                _profilingStatus.fontColor = Colors.WHITE;
                _profilingStatus.x = LEFT_MARGIN;
                _profilingStatus.y = 20;
            }
            return _profileTab;
        }

        /**
         * make a status display widget
         */
        protected function createStatusTextField():TextField
        {
            var tf:TextField = internalGameObjects.add(TextField) as TextField;
            tf.fontColor = Colors.WHITE;
            tf.x = LEFT_MARGIN;
            tf.y = 60;
            return tf;
        }

        /**
         * clear the log
         */
        public function clearLog(e:Event = null):void
        {
            if (_logTextField)
            {
                _logTextField.setText("");
            }
        }

        /**
         * add to the log
         */
        public function log(text:String):void
        {
            if (_logTextField)
            {
                // TODO:  make into a ginormous scrolling thing someday
                _logTextField.setText(_logTextField.getText()+"\n"+text);
            }
        }

        /**
         *
         */
        override public function update(deltaTime:Number):void
        {
            super.update(deltaTime);
            var text:String, c:Object, key:String, ll:LinkedList;

            if (visible)
            {
                if (_poolStatus && _poolStatus.visible)
                {
                    text = "object pool info\n'new's:  " + (ObjectPool.numNews + ObjectPool.numBitmapDataNews + ObjectPool.numLinkedListNodeNews );
                    for (c in ObjectPool.pools)
                    {
                        ll = ObjectPool.pools[c];
                        text += "\n" + getQualifiedClassName(c) + ":  " + ll.count;
                    }
                    for (key in ObjectPool.bitmapDataPool)
                    {
                        ll = ObjectPool.bitmapDataPool[key];
                        text += "\nbitmapData " + key + ":  " + ll.count;
                    }
                    text += "\nlinkedListNode:  " + ObjectPool.linkedListNodePool.length;

                    _poolStatus.setText(text);
                }
                if (_reusablesStatus && _reusablesStatus.visible)
                {
                    text = "reusable game objects info\n'new's:  " + GameObjectHolder.numNews;
                    for (c in GameObjectHolder.reusables)
                    {
                        var a:Array = GameObjectHolder.reusables[c];
                        text += "\n" + getQualifiedClassName(c) + ":  " + a.length;
                    }
                    _reusablesStatus.setText(text);
                }
                if (_profilingStatus && _profilingStatus.visible)
                {
                    text = "Profiling info\n";
                    text += StringUtil.fixedWidth("name", 25);
                    text += StringUtil.fixedWidth("times", 8);
                    text += StringUtil.fixedWidth("avg", 6);
                    text += StringUtil.fixedWidth("min", 4);
                    text += StringUtil.fixedWidth("max", 4);
                    text += StringUtil.fixedWidth("total", 6);
                    for (var sectionName:String in Profiler.sections)
                    {
                        var ps:ProfilerSection = Profiler.sections[sectionName];
                        text += "\n" + StringUtil.fixedWidth(sectionName, 25);
                        text += StringUtil.fixedWidth(ps.times.toString(), 8);
                        text += StringUtil.fixedWidth(ps.avg.toFixed(3), 6);
                        text += StringUtil.fixedWidth(ps.min.toString(), 4);
                        text += StringUtil.fixedWidth(ps.max.toString(), 4);
                        text += StringUtil.fixedWidth(ps.total.toString(), 6);
                    }
                    _profilingStatus.setText(text);
                }
            }
        }

        /**
         * cleanup
         */
        override public function shutdown():void
        {
            super.shutdown();
            game = null;
        }

        /**
         * draw the hitboxes for the objects in the layer
         */
        protected function drawHitboxesInLayer(layer:GameObjectHolder):void
        {
            layer.forEach(drawHitboxesForObject);
        }

        /**
         * draw the hitboxes for the object
         */
        protected function drawHitboxesForObject(go:GameObject):void
        {
            var canvas:Canvas = _hitboxLayer.canvas,
                hitboxes:LinkedList = go.getHitboxes(),
                node:LinkedListNode = hitboxes.first,
                hitbox:HitBox;
            while (node)
            {
                hitbox = node.value.clone();
                if ( (hitbox.width > 0) && (hitbox.height > 0) )
                {
                    hitbox.offset(go.x, go.y);
                    canvas.fillRectBlended(hitbox, (0xFFFFFF & hitbox.color) | 0x7F000000);
                }
                node = node.next;
            }
        }
        override public function preDraw(deltaTime:Number):void
        {
            super.preDraw(deltaTime);
            if (visible)
            {
                if (_hitboxLayer.visible)
                {
                    var canvas:Canvas = _hitboxLayer.canvas;
                    canvas.clear();
                    game.layers.layers.forEach(drawHitboxesInLayer);
                }
            }
        }
    }
}
