package stego.gameobject
{
    import flash.display.BitmapData;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.collections.LinkedList;
    import stego.collections.LinkedListNode;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.gameobject.TileData;
    import stego.util.collision.CollisionInfo;
    import stego.util.collision.CollisionUtil;
    import stego.util.collision.HitBox;
    import stego.util.ObjectPool;

    /**
     * GameObject for tile maps
     * Set the five properties, then start calling setTile.
     * @author Mike Meyer
     */
    public class TileMap extends CanvasGameObject
    {
        /** image data from which to get the images for each tile. */
        public function get tilesetImage():BitmapData { return _tilesetImage; }
        public function set tilesetImage(value:BitmapData):void
        {
            _tilesetImage = value;
            if (null == _tilesetImage)
            {
                _tilesetColumns = 0;
                _tilesetRows = 0;
                _maxTileIndex = 0;
                _needsTilesRedrawnToBuffer = true;
            }
            else
            {
                _tilesetColumns = _tilesetImage.width / _tileWidth;
                _tilesetRows = _tilesetImage.height / _tileHeight;
                _maxTileIndex = _tilesetColumns * _tilesetRows;
                _needsTilesRedrawnToBuffer = true;
            }
        }
        protected var _tilesetImage:BitmapData;

        /** width of each tile. */
        public function get tileWidth():uint { return _tileWidth; }
        public function set tileWidth(value:uint):void
        {
            _tileWidth = value;
            _tilesetColumns = _tilesetImage.width / _tileWidth;
            _maxTileIndex = _tilesetColumns * _tilesetRows;
            _needsTilesRedrawnToBuffer = true;
        }
        protected var _tileWidth:uint;

        /** height of each tile. */
        public function get tileHeight():uint { return _tileHeight; }
        public function set tileHeight(value:uint):void
        {
            _tileHeight = value;
            _tilesetRows = _tilesetImage.height / _tileHeight;
            _maxTileIndex = _tilesetColumns * _tilesetRows;
            _needsTilesRedrawnToBuffer = true;
        }
        protected var _tileHeight:uint;

        /** width of the entire tileset image */
        override public function set width(value:Number):void
        {
            if (_canvas.width > value)
            {
                _needsTilesRedrawnToBuffer = true;
            }
            super.width = value;
        }

        /** height of the entire tileset image */
        override public function set height(value:Number):void
        {
            if (_canvas.height > value)
            {
                _needsTilesRedrawnToBuffer = true;
            }
            super.height = value;
        }

        /** whether to check for collisions against the edges of the tileset */
        public var checkMapEdgeCollisions:Boolean = false;

        /** 2d array of tile info */
        protected var _tileData:Array;

        /** how many columns are in the tileset source image */
        protected var _tilesetColumns:uint;
        /** how many rows are in the tileset source image */
        protected var _tilesetRows:uint;
        /** highest valid tile index */
        protected var _maxTileIndex:uint;

        /** whether the currently drawn graphics are out of date and need updating */
        protected var _needsTilesRedrawnToBuffer:Boolean;

        /**
         * optimization. copyTileIntoBuffer needs a rect and getting/creating
         * one would be a significant portion of the function's cost.
         * it is called a lot so let's give it a dedicated tempRect
         */
        protected static var _tempRect:Rectangle = new Rectangle();

        /**
         * initial setup and preparation for reuse
         */
        override public function init(game:GameInterface):void
        {
            super.init(game);
            _maxTileIndex = 0;
            _tileData = [];
            _tilesetColumns = 0;
            _tilesetRows = 0;
            _tileHeight = 16;
            _tileWidth = 16;
            _tilesetImage = null;
            _needsTilesRedrawnToBuffer = true;
        }

        /**
         * called after all the final positions have been figured
         * just before drawing
         */
        override public function preDraw(deltaTime:Number):void
        {
            super.preDraw(deltaTime);

            if (_needsTilesRedrawnToBuffer)
            {
                updateTileMap();
            }
        }

        /**
         * clear the whole tilemap
         */
        public function clear():void
        {
            _tileData = [];
            _canvas.clear();
            _needsTilesRedrawnToBuffer = false;
        }

        /**
         * clear a tile
         */
        public function clearTile(column:uint, row:uint):void
        {
            var tileData:TileData = getTileData(column, row, false);
            if (tileData)
            {
                setTileInternal(tileData, -1);
            }
        }

        /** clear a block of tiles */
        public function clearTiles( startRow:uint,
                                    startCol:uint,
                                    endRow:uint,
                                    endCol:uint ):void
        {
            var rowData:Array,
                tileData:TileData;
            for (var row:uint = startRow; row <= endRow; ++row)
            {
                rowData = _tileData[row];
                if (null != rowData)
                {
                    for (var col:uint = startCol; col <= endCol; ++col)
                    {
                        tileData = rowData[col];
                        if (null != tileData)
                        {
                            setTileInternal(tileData, -1);
                        }
                    }
                }
            }
        }

        /**
         * set which index to use for a tile. -1 for empty.
         */
        public function setTile(column:uint, row:uint, index:int, userData:Object = null):void
        {
            setTileInternal(getTileData(column, row, true), index, userData);
        }

        /**
         * Helper for setTile. actually performs the work after setTile finds/creates the tile
         * @param	tileData
         * @param	index
         * @param   userData (optional)
         */
        protected function setTileInternal(tileData:TileData, index:int, userData:Object = ""):void
        {
            tileData.index = index;
            tileData.userData = userData;

            if (_tilesetImage)
            {
                copyTileIntoBuffer(tileData);
            }
        }

        /**
         * get the data object to go with the tile at the given location.
         * @param   createIfNeeded  If true, this will create the tile if it doesn't exist yet
         */
        public function getTileData(column:uint, row:uint, createIfNeeded:Boolean=false):TileData
        {
            var rowData:Array = _tileData[row];
            if (!rowData)
            {
                if (!createIfNeeded) { return null; }
                rowData = _tileData[row] = [];
            }
            var tileData:TileData = rowData[column];
            if (!tileData)
            {
                if (!createIfNeeded) { return null; }
                tileData = rowData[column] = newTileData(-1, row, column);
            }
            return tileData;
        }

        /**
         * to make it easier to override with custom tiledata
         */
        protected function newTileData(index:int, row:int, column:int):TileData
        {
            return new TileData(index, row, column);
        }

        /**
         * Copy a tile image into a spot on the buffer
         * does not check that it is safe to do so
         * @param	tileData    tileData for the tile we are rendering
         */
        protected function copyTileIntoBuffer(tileData:TileData):void
        {
            var tempPoint:Point = ObjectPool.tempPoint;
            var index:int = tileData.index;
            var column:Number = tileData.col;
            var row:Number = tileData.row;

            if (index >= 0)
            {
                _tempRect.x = (index % _tilesetColumns) * _tileWidth;
                _tempRect.y = uint(index / _tilesetColumns) * _tileHeight;
                _tempRect.width = _tileWidth;
                _tempRect.height = _tileHeight;
                tempPoint.x = column * _tileWidth;
                tempPoint.y = row * _tileHeight;
                internalCopyTileIntoBuffer(tileData, _tempRect, tempPoint);
            }
            else
            {
                _tempRect.x = column * _tileWidth;
                _tempRect.y = row * _tileHeight;
                _tempRect.width = _tileWidth;
                _tempRect.height = _tileHeight;
                internalClearTileInBuffer(tileData, _tempRect);
            }
        }

        /**
         * helper for copyTileIntoBuffer.  extracted for easier overriding
         * @param	tileData        tileData for the tile we are rendering
         * @param	sourceRect      source area of the tile image
         * @param   destPoint       where in the buffer to render it at
         */
        protected function internalCopyTileIntoBuffer( tileData:TileData,
                                                       sourceRect:Rectangle,
                                                       destPoint:Point ):void
        {
            _canvas.copyPixels( _tilesetImage,
                                sourceRect,
                                destPoint,
                                true );
        }

        /**
         * helper for copyTileIntoBuffer.  extracted for easier overriding
         * @param	tileData    tileData for the tile we are rendering
         * @param	rect        area in the buffer for that tile
         */
        protected function internalClearTileInBuffer( tileData:TileData,
                                                      rect:Rectangle ):void
        {
            _canvas.fillRect(rect,0);
        }

        /**
         * update the big image used for the tile map graphics
         */
        protected function updateTileMap():void
        {
            for (var row:uint = 0; row < _tileData.length; ++row)
            {
                var rowData:Array = _tileData[row];
                if (!rowData) { continue; }
                for (var column:uint = 0; column < rowData.length; ++column)
                {
                    var tileData:TileData = rowData[column];
                    if (!tileData ||  tileData.index<0) { continue; }
                    copyTileIntoBuffer(tileData);
                }
            }
            _needsTilesRedrawnToBuffer = false;
        }

        /**
         * clean up
         */
        override public function shutdown():void
        {
            super.shutdown();
            _tilesetImage = null;
            _tileData = null;
        }

        /**
         * detect and react to collisions with a regular GameObject
         * @param   go          object to collide against the tilemap
         * @param   callback    for each tile that go collides with, this will
         *                      get called and sent the tileData object (that
         *                      you can modify), tileMap, game object, and
         *                      collisionInfo.  so it should look like
         *                      function tileCollisionCallback(tileData:TileData, tileMap:TileMap, go:GameObject, ci:CollisionInfo):void
         */
        public function collide( go:GameObject,
                                 deltaTime:Number,
                                 callback:Function = null ):void
        {
            var hitboxes:LinkedList = go.getHitboxes(),
                hitbox:HitBox,
                node:LinkedListNode = hitboxes.first;

            var offsetLeft:Number, // distance from the left edge of the tilemap to the left edge of the object
                offsetTop:Number, // distance from the top of the tilemap to the top of the object
                offsetRight:Number, // distance from the left edge of the tilemap to the right edge of the object
                offsetBottom:Number; // distance from the top of the tilemap to the bottom of the object

            while (node)
            {
                hitbox = node.value;

                offsetLeft = (go.x + hitbox.x - x);
                offsetTop = (go.y + hitbox.y - y);
                offsetRight = offsetLeft + hitbox.width;
                offsetBottom = offsetTop + hitbox.height;

                if ( (offsetLeft   < width)  &&
                     (offsetTop    < height) &&
                     (offsetRight  > 0.0)    &&
                     (offsetBottom > 0.0) )
                {
                    var firstRow:uint = (offsetTop < 0.0) ? 0 : uint(offsetTop / _tileHeight);
                    var lastRow:uint = (offsetBottom < 0.0) ? 0 : uint(offsetBottom / _tileHeight);
                    lastRow = (offsetBottom == lastRow) ? lastRow : lastRow + 1;

                    var firstCol:uint = (offsetLeft < 0.0) ? 0 : uint(offsetLeft / _tileWidth);
                    var lastCol:uint = (offsetRight < 0.0) ? 0 : uint(offsetRight / _tileWidth);
                    lastCol = (offsetRight == lastCol) ? lastCol : lastCol + 1;

                    var tileRect:Rectangle = ObjectPool.getObject(Rectangle),
                        row:uint,
                        col:uint,
                        row_i:uint,
                        row_i_max:uint = lastRow-firstRow,
                        col_i:uint,
                        col_i_max:uint = lastCol-firstCol,
                        rowData:Array,
                        tileData:TileData,
                        ci:CollisionInfo,
                        // look at its motion and first check the tiles that it would hit earliest;
                        goRight:Boolean = (go.dx >= 0),
                        goDown:Boolean = (go.dy >= 0);

                    for (row_i=0; row_i<=row_i_max; ++row_i)
                    {
                        row = goDown ? firstRow + row_i : lastRow - row_i;
                        rowData = _tileData[row];
                        if (!rowData) { continue; }
                        for (col_i=0; col_i <= col_i_max; ++col_i)
                        {
                            col = goRight ? firstCol+col_i : lastCol-col_i;
                            tileData = rowData[col];
                            if ( !tileData || (tileData.index < 0) ) { continue; }

                            collideTile(tileData, go, hitbox, deltaTime, callback);
                        }
                    }

                    // check map edges (optionally)
                    if (checkMapEdgeCollisions)
                    {
                        ci = ObjectPool.getObject(CollisionInfo);
                        if (offsetLeft < 0)
                        {
                            ci.t = 0;
                            ci.normal.x = 1;
                            ci.normal.y = 0;
                            ci.location.x = x;
                            ci.location.y = go.y + 0.5 * (hitbox.top + hitbox.bottom);

                            // let the character know he hit something
                            ci.hitbox1 = hitbox;
                            go.collisionOccurred(ci, this);
                            ci.normal.x *= -1;
                            ci.normal.y *= -1;

                            this.collisionOccurred(ci, go);
                        }
                        if (offsetRight > width)
                        {
                            ci.t = 0;
                            ci.normal.x = -1;
                            ci.normal.y = 0;
                            ci.location.x = width;
                            ci.location.y = go.y + 0.5 * (hitbox.top + hitbox.bottom);

                            // let the character know he hit something
                            ci.hitbox1 = hitbox;
                            go.collisionOccurred(ci, this);
                            ci.normal.x *= -1;
                            ci.normal.y *= -1;

                            this.collisionOccurred(ci, go);
                        }
                        if (offsetTop < 0)
                        {
                            ci.t = 0;
                            ci.normal.x = 0;
                            ci.normal.y = 1;
                            ci.location.x = go.x + 0.5 * (hitbox.left + hitbox.right);
                            ci.location.y = y;

                            // let the character know he hit something
                            ci.hitbox1 = hitbox;
                            go.collisionOccurred(ci, this);
                            ci.normal.x *= -1;
                            ci.normal.y *= -1;

                            this.collisionOccurred(ci, go);
                        }
                        if (offsetBottom > height)
                        {
                            ci.t = 0;
                            ci.normal.x = 0;
                            ci.normal.y = -1;
                            ci.location.x = go.x + 0.5 * (hitbox.left + hitbox.right);
                            ci.location.y = height;

                            // let the character know he hit something
                            ci.hitbox1 = hitbox;
                            go.collisionOccurred(ci, this);
                            ci.normal.x *= -1;
                            ci.normal.y *= -1;

                            this.collisionOccurred(ci, go);
                        }
                        ObjectPool.returnObject(CollisionInfo, ci);
                    }
                    ObjectPool.returnObject(Rectangle, tileRect);
                }

                node = node.next;
            }
        }

        /**
         * do a collision check for one hitbox/tile pair
         */
        protected function collideTile( tileData:TileData,
                                        go:GameObject,
                                        hitbox:HitBox,
                                        deltaTime:Number,
                                        callback:Function = null ):void
        {
            var offsetLeft:Number = (hitbox.x - x), // distance from the left edge of the tilemap to the left edge of the object
                offsetTop:Number = (hitbox.y - y), // distance from the top of the tilemap to the top of the object
                offsetRight:Number = offsetLeft + hitbox.width, // distance from the left edge of the tilemap to the right edge of the object
                offsetBottom:Number = offsetTop + hitbox.height, // distance from the top of the tilemap to the bottom of the object
                ci:CollisionInfo;
            var tileHb:HitBox = ObjectPool.getObject(HitBox);
            fillOutHitboxForTile(tileHb, tileData);

            ci = CollisionUtil.checkHitboxCollision(hitbox, go, tileHb, this, deltaTime);
            if (ci)
            {

                // let the character know he hit something
                go.collisionOccurred(ci, this);
                ci.reverse();
                this.collisionOccurred(ci, go);
                ci.reverse();

                // call the callack, if we were given one
                if (null != callback)
                {
                    callback(tileData, this, go, ci);
                }
            }
            ObjectPool.returnObject(HitBox, tileHb);
        }

        /** set the values in tileHb to reflect the hitbox for the given tile */
        protected function fillOutHitboxForTile(tileHb:HitBox, tileData:TileData):void
        {
            tileHb.x = (tileData.col * _tileWidth);
            tileHb.y = (tileData.row * _tileHeight);
            tileHb.width = _tileWidth;
            tileHb.height = _tileHeight;
            tileHb.userData = tileData.userData;
        }

        /**
         * tilemaps never move in response to collisions
         */
        override public function collisionOccurred(ci:CollisionInfo, object:GameObject):void
        {
        }

        /**
         * update the hitboxes array
         * Redoes it from scratch every frame using getBounds!
         * Override to be smarter and/or turn off autoUpdateHitboxes!
         */
        override protected function calculateHitboxes():void
        {
            var hb:HitBox;
            if (!hitboxes) { hitboxes = new LinkedList; }
        }

        /** get how many tiles wide the tilemap is */
        public function getNumColumns():uint
        {
            return width / _tileWidth;
        }

        /** get how many tiles tall the tilemap is */
        public function getNumRows():uint
        {
            return height / _tileHeight;
        }
    }
}

