/**
 * Wrapper for Flash text fields to add convenient methods and
 * automatically set good defaults for this game.
 */
package stego.ui
{
    import flash.text.TextField;
    import flash.text.TextFieldAutoSize;
    import flash.text.TextFormat;
    import stego.GameInterface;
    import stego.GameObject;
    import stego.constants.Colors;

    public class TextField extends GameObject
    {
        [Embed(source = '../assets/Fonts/ankylosaurus.ttf', fontName="ankylosaurus")]
        private static const _font:Class;

        /**
         * get/set the default font to use for textfields
         */
        public static var defaultFont:String = "ankylosaurus";

        /**
         * get/set the default font size to use for textfields
         */
        public static var defaultFontSize:Number = 8;

        /**
         * get/set the default color to use for textfields
         */
        public static var defaultFontColor:Number = Colors.WHITE_ENOUGH;

        /**
         * actual textfield to wrap
         */
        protected var textField:flash.text.TextField;

        public static const HALIGN_LEFT:uint       = 0;
        public static const HALIGN_CENTER:uint     = 1;
        public static const HALIGN_RIGHT:uint      = 2;
        public static const VALIGN_TOP:uint        = 3;
        public static const VALIGN_MIDDLE:uint     = 4;
        public static const VALIGN_BOTTOM:uint     = 5;

        private static const DEFAULT_WORDWRAP_WIDTH:Number = 200;

        private var _hAlign:uint;
        public function get hAlign():uint { return _hAlign; }
        public function set hAlign(value:uint):void
        {
            _hAlign = value;
            updateHAlign();
        }
        private var _vAlign:uint;
        public function get vAlign():uint { return _vAlign; }
        public function set vAlign(value:uint):void
        {
            _vAlign = value;
            updateVAlign();
        }

        public function get wordWrap():Boolean { return textField.wordWrap; }
        public function set wordWrap(value:Boolean ):void { textField.wordWrap = value; resetSize(); }

        public function get wordWrapWidth():Number { return textField.width; }
        public function set wordWrapWidth(value:Number):void { textField.width = value; resetSize(); }

        public function TextField():void
        {
            super();

            if (null == textField)
            {
                textField = new flash.text.TextField();
                textField.autoSize = TextFieldAutoSize.LEFT;
                textField.multiline = true;
                this.addChild(textField);
            }
            resetDefaults();
        }

        /**
         * this function called for both initial setup and to re-init for reuse
         */
        public override function init(game:GameInterface):void
        {
            super.init(game);
            resetDefaults();
        }

        /**
         * Change the displayed text.
         */
        public function setText(desiredText:String):void
        {
            textField.text = desiredText;
            resetSize();
        }

        /**
         * Change the displayed text.
         */
        public function setHtmlText(desiredHtmlText:String):void
        {
            textField.htmlText = desiredHtmlText;
            resetSize();
        }

        /**
         * Get the displayed text.
         */
        public function getText():String
        {
            return textField.text;
        }

        /**
         * Change the font.
         */
        public function setFont(font:String):void
        {
            var fmt:TextFormat = textField.getTextFormat();
            fmt.font = font;
            textField.defaultTextFormat = fmt;
            textField.setTextFormat(fmt);

            resetSize();
        }

        /**
         * Change the displayed font size.
         */
        public function setFontSize(desiredSize:Number):void
        {
            var fmt:TextFormat = textField.getTextFormat();
            fmt.size = desiredSize;
            textField.defaultTextFormat = fmt;
            textField.setTextFormat(fmt);

            resetSize();
        }

        /**
         * Get the displayed font size.
         */
        public function getFontSize():Number
        {
            return textField.getTextFormat().size as Number;
        }

        /**
         * used internally to keep .width and .height sensible
         */
        protected function resetSize():void
        {
            updateHAlign();
            updateVAlign();
        }

        /**
         * center the textfield on some point.  CALL SETTEXT() *FIRST*
         */
        public function centerOn(center_x:Number, center_y:Number):void
        {
            x = center_x;
            y = center_y;
            hAlign = HALIGN_CENTER;
            vAlign = VALIGN_MIDDLE;
        }

        /**
         * reset to defaults.  This thing will likely get re-used!
         */
        public function resetDefaults():void
        {
            textField.x = 0;
            textField.y = 0;
            textField.selectable = false;
            textField.embedFonts = true;

            var format:TextFormat = new TextFormat();
            format.font = defaultFont;
            format.size = defaultFontSize;
            textField.defaultTextFormat = format;
            textField.setTextFormat(format);
            textField.textColor = defaultFontColor;
            textField.text = "";

            hAlign = HALIGN_LEFT;
            vAlign = VALIGN_TOP;
            wordWrap = false;
            //WordWrapWidth = DEFAULT_WORDWRAP_WIDTH;
            resetSize();
        }

        /**
         * Change the color of the displayed text
         */
        public function set fontColor(color:uint):void
        {
            textField.textColor = color;
        }

        /**
         * the color of the displayed text
         */
        public function get fontColor():uint
        {
            return textField.textColor;
        }

        private function updateHAlign():void
        {
            switch (hAlign)
            {
                case HALIGN_LEFT:
                    textField.x = 0.0;
                    break;
                case HALIGN_CENTER:
                    textField.x = -0.5 * textField.textWidth;
                    break;
                case HALIGN_RIGHT:
                    textField.x = -1.0 * textField.textWidth;
                    break;
                default:
                    throw new Error("unrecognized h alignment");
                    break;
            }
        }
        private function updateVAlign():void
        {
            switch (vAlign)
            {
                case VALIGN_TOP:
                    textField.y = 0;
                    break;
                case VALIGN_MIDDLE:
                    // the way textfield layout works, centering veritcally
                    // looks like centering way too low to me.  Fudging it
                    // upward to look better.
                    textField.y = -0.6 * textField.textHeight;
                    break;
                case VALIGN_BOTTOM:
                    textField.y = -textField.textHeight;
                    break;
                default:
                    throw new Error("unrecognized v alignment");
                    break;
            }
        }
    }
}
