﻿package stego.util
{
    import flash.geom.Matrix;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    /**
     * Math-related utility functions
     * @author Mike Meyer
     */
    public class MathUtil
    {
        /** multiply by this to go from degrees to radians */
        public static const deg2Rad:Number = Math.PI / 180.0;

        /** multiply by this to go from radians to degrees */
        public static const rad2Deg:Number = 180.0 / Math.PI;

        /** avoid unnecessary 'new's */
        private static var _tempPoint1:Point = new Point();
        private static var _tempPoint2:Point = new Point();

        /**
         * randomly choose a number in the range [min, max)
         */
        public static function randomRange(min:Number, max:Number):Number
        {
            return min + (Math.random() * (max-min))
        }

        /**
         * Do a linear interpolation between two values.  t in [0,1]
         */
        public static function lerp(a:Number, b:Number, t:Number):Number
        {
            return a + (t*(b-a));
        }

        /**
         * inverse of lerp, given a, b, and the result, figure out what t was
         */
        public static function unlerp(value:Number, a:Number, b:Number):Number
        {
            return (value - a) / (b - a);
        }

        /**
         * Do a linear mapping between two ranges and apply that mapping to x
         */
        public static function remap( x:Number,
                                      fromRangeMin:Number,
                                      fromRangeMax:Number,
                                      toRangeMin:Number,
                                      toRangeMax:Number ):Number
        {
            return toRangeMin + ( (x - fromRangeMin) / (fromRangeMax - fromRangeMin) * (toRangeMax - toRangeMin) )
        }

        /**
         * Keep a value within the range
         */
        public static function clamp(value:Number, min:Number, max:Number):Number
        {
            return Math.max( min, Math.min( max, value ) );
        }

        /**
         * Wrap a value around to keep it within a range
         */
        public static function wrap(value:Number, min:Number, max:Number):Number
        {
            if ( (max - min) == 0 )
            {
                return min;
            }

            // deal with the edge cases
            if (max < min)
            {
                var temp:Number = max;
                max = min;
                min = temp;
            }

            var wrapped:Number = ( (value-min) % (max - min)) + min;
            if (wrapped < min) { wrapped += (max - min); }
            return wrapped;
        }

        /**
         * bounce a value back and forth to keep it within a range
         */
        public static function pingpong( value:Number,
                                         min:Number,
                                         max:Number,
                                         easing:Function = null ):Number
        {
            var t:Number = wrap( (value - min) / (max - min), 0.0, 2.0);
            if (t <= 1.0)
            {
                return lerp( min,
                             max,
                             (null == easing) ? t : easing(t));
            }
            else
            {
                t -= 1.0;
                return lerp( max,
                             min,
                             (null == easing) ? t : easing(t) );
            }
        }

        /**
         * given a rect, returns a new point telling you where its
         * center is
         */
        public static function getRectCenter(rect:Rectangle):Point
        {
            return new Point(
                0.5*(rect.left+rect.right),
                0.5*(rect.top+rect.bottom) );
        }

        /**
         * given an x and y, determines the slope of a line
         * through (0,0) and that point.
         * MAY RETURN INFINITY (or negative infinity)
         * if you give it (0,0), it's obviously not going
         * to be able to give you a slope
         */
        public static function getSlope(x:Number, y:Number):Number
        {
            if (0.0 != x)
            {
                return (y / x);
            }
            else if (0.0 > y)
            {
                return Number.NEGATIVE_INFINITY;
            }
            else
            {
                return Number.POSITIVE_INFINITY;
            }
        }

        /**
         * returns the "sign" of a number
         * n > 0  ==> 1
         * n < 0  ==> -1
         * n = 0  ==> 0
         */
        public static function sign(n:Number):Number
        {
            if (n > 0) return 1;
            if (n < 0) return -1;
            return 0;
        }

        /**
         * return true if the numbers are both negative or both positive
         * if either is zero, this returns false
         */
        public static function sameSign(x:Number, y:Number):Boolean
        {
            return ((x*y)>0.0);
        }

        /**
         * return true if one number is negative and the other positive
         * if either is zero, this returns false
         */
        public static function oppositeSign(x:Number, y:Number):Boolean
        {
            return ((x*y)<0.0);
        }

        /**
         * Scale a point/vector and return it (leaves the original alone)
         */
        public static function scaleVector(v:Point, scale:Number):Point
        {
            return new Point(v.x * scale, v.y * scale);
        }

        /**
         * if the vector is longer than length, shorten it to length and
         * return it (leaves the original alone)
         */
        public static function capVectorLength(v:Point, length:Number):Point
        {
            var curLength:Number = Math.sqrt( (v.x * v.x) + (v.y * v.y) );
            if (curLength > length)
            {
                return scaleVector(v, length / curLength);
            }
            else
            {
                return v.clone();
            }
        }

        /**
         * move at most maxDist> units from <a> in the direction of <b>,
         * stopping at <b> if it's enough to take us there
         */
        public static function moveToward(a:Number, b:Number, maxDist:Number):Number
        {
            if ( Math.abs(b - a) < maxDist )
            {
                return b;
            }
            else if (b > a)
            {
                return a + maxDist;
            }
            else
            {
                return a - maxDist;
            }
        }

        /**
         * move at most <maxDist> units from <a> in the direction of <b>,
         * stopping at <b> if it's enough to take us there
         */
        public static function moveToward2d(a:Point, b:Point, maxDist:Number):Point
        {
            var distance:Number = Point.distance(a, b);

            if (distance < maxDist)
            {
                return b;
            }
            else
            {
                var delta:Point = b.subtract(a);
                delta.normalize(maxDist);

                return a.add(delta);
            }
        }

        /**
         * rotate at most <maxRotation> units from <a> in the direction of <b>,
         * stopping at <b> if it's enough to take us there
         */
        public static function rotateToward(a:Number, b:Number, maxRotation:Number, radians:Boolean = false):Number
        {
            var wrapMax:Number = radians ? Math.PI : 180;
            var wrapMin:Number = -wrapMax;

            var difference:Number = MathUtil.wrap( b - a, wrapMin, wrapMax );

            if (Math.abs(difference) < maxRotation)
            {
                return b;
            }
            else
            {
                return (difference > 0) ? (a + maxRotation): (a - maxRotation);
            }
        }

        /**
         * round a to the nearest multiple of b
         */
        public static function roundToNearestMultiple(a:Number, b:Number):Number
        {
            return int((a / b)+0.5) * b;
        }

        /**
         * round DOWN to the nearest multiple of b
         */
        public static function roundDownToNearestMultiple(a:Number, b:Number):Number
        {
            return int(a / b) * b;
        }

        /**
         * round UP to the nearest multiple of b
         */
        public static function roundUpToNearestMultiple(a:Number, b:Number):Number
        {
            var roundedDown:Number = int(a / b) * b;
            if ( roundedDown == a ) { return roundedDown;  }
            else if ( a > 0 ) { return roundedDown + b; }
            else  { return roundedDown; }
        }

        /**
         * distance between two locations in 2d
         */
        public static function distance2d( x1:Number,
                                           y1:Number,
                                           x2:Number,
                                           y2:Number ):Number
        {
            var dx:Number = x2 - x1;
            var dy:Number = y2 - y1;
            return Math.sqrt( (dx * dx) + (dy * dy) );
        }

        /**
         * sets up the given matrix to represent the transform specified by the other args
         * @param	matrix      matrix to modify
         * @param	originX     offset from x that should be considered the center for the other transforms
         * @param	originY     offset from y that should be considered the center for the other transforms
         * @param	x           x translation to apply after scale and rotation
         * @param	y           y translation to apply after scale and rotation
         * @param	rotation    rotation to apply after scale and before translation. in radians
         * @param	xScale      how much to scale along the x axis before rotation and translation
         * @param	yScale      how much to scale along the y axis before rotation and translation
         */
        public static function setupTransform( matrix:Matrix,
                                               originX:Number = 0,
                                               originY:Number = 0,
                                               x:Number = 0,
                                               y:Number = 0,
                                               rotation:Number = 0,
                                               xScale:Number = 1,
                                               yScale:Number = 1 ):void
        {
            matrix.a = xScale;
            matrix.b = 0;
            matrix.c = 0;
            matrix.d = yScale;
            matrix.tx = -originX * xScale;
            matrix.ty = -originY * yScale;
            matrix.rotate(rotation);
            matrix.tx += x;
            matrix.ty += y;
        }

        /**
         * transform the rectangle by the matrix.
         * if outputRect is not null, that rect will be changed into the tranformed rect (and returned)
         * otherwise, a new rect is returned
         */
        public static function transformRect(rect:Rectangle, matrix:Matrix, outputRect:Rectangle):Rectangle
        {
            if (null == outputRect)
            {
                outputRect = new Rectangle();
            }

            _tempPoint1 = matrix.transformPoint(rect.topLeft);
            _tempPoint2 = matrix.transformPoint(rect.bottomRight);

            outputRect.x = Math.min(_tempPoint1.x, _tempPoint2.x);
            outputRect.y = Math.min(_tempPoint1.y, _tempPoint2.y);
            outputRect.width = Math.abs(_tempPoint1.x - _tempPoint2.x);
            outputRect.height = Math.abs(_tempPoint1.y - _tempPoint2.y);

            return outputRect;
        }

        /**
         * round up to the next highest power of 2
         */
        public static function roundUpToPowerOf2(v:int):int
        {
            --v;
            v |= v >> 1;
            v |= v >> 2;
            v |= v >> 4;
            v |= v >> 8;
            v |= v >> 16;
            return ++v;
        }
    }
}
