﻿package stego.util.collision
{
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.collections.LinkedList;
    import stego.collections.LinkedListNode;
    import stego.GameObject;
    import stego.util.collision.CollisionInfo;

    /**
     * handy functions for doing collision checks
     * @author Mike Meyer
     */
    public class CollisionUtil
    {
        // the farthest (relative) distance we want to go between collision tests
        private static const COLLISION_STEP_SIZE_MAX:Number = 4;

        /** avoid new'ing */
        protected static var _tempHitBox1:HitBox = new HitBox();
        protected static var _tempHitBox2:HitBox = new HitBox();
        protected static var _tempPoint1:Point = new Point();
        protected static var _tempPoint2:Point = new Point();
        protected static var _tempPoint3:Point = new Point();

        /**
         * check to see if the hitboxes for the two obejcts overlap
         * @param	objectA - first object in the collision-check pair
         * @param	objectB - second object in the collision-check pair
         * @return  True if there was a collision.
         */
        public static function objectsOverlap( objectA:GameObject,
                                               objectB:GameObject ):Boolean
        {
            var collided:Boolean = false,
                aHitboxes:LinkedList = objectA.getHitboxes(),
                aNode:LinkedListNode = aHitboxes.first,
                objectAHitBox:HitBox,
                bHitboxes:LinkedList = objectB.getHitboxes(),
                bNode:LinkedListNode = bHitboxes.first,
                objectBHitBox:HitBox;
            while (aNode)
            {
                objectAHitBox = aNode.value;
                aNode = aNode.next;
                while (bNode)
                {
                    objectBHitBox = bNode.value;
                    bNode = bNode.next;
                    objectAHitBox.offset(objectA.x, objectA.y);
                    objectBHitBox.offset(objectB.x, objectB.y);
                    if (objectAHitBox.intersects(objectBHitBox))
                    {
                        collided = true;
                        break;
                    }
                }
                if (collided)
                {
                    break;
                }
            }

            return collided;
        }

        /**
         * check for collisions between two moving GameObjects
         * returns null if they did not collide
         * @param	collideCallback - function to call in the event of a collision
         * @param	objectA - first object in the collision-check pair
         * @param	objectB - second object in the collision-check pair
         * @param	deltaTime - length of the time step in which we are checking collisions
         */
        public static function checkAabbCollision( collideCallback:Function,
                                                   objectA:GameObject,
                                                   objectB:GameObject,
                                                   deltaTime:Number ):void
        {
            var ci:CollisionInfo,
                aHitboxes:LinkedList = objectA.getHitboxes(),
                aNode:LinkedListNode = aHitboxes.first,
                objectAHitBox:HitBox,
                bHitboxes:LinkedList = objectB.getHitboxes(),
                bNode:LinkedListNode = bHitboxes.first,
                objectBHitBox:HitBox;

            while (aNode)
            {
                objectAHitBox = aNode.value;
                aNode = aNode.next;
                bNode = bHitboxes.first;
                while (bNode)
                {
                    objectBHitBox = bNode.value;
                    bNode = bNode.next;
                    ci = checkHitboxCollision( objectAHitBox,
                                               objectA,
                                               objectBHitBox,
                                               objectB,
                                               deltaTime );

                    if (null != ci)
                    {
                        collideCallback( ci, objectA, objectB );
                    }
                }
            }
        }

        /**
         * check for collisions between the specified hitboxes of the specified objects
         * returns null if they did not collide
         * @param   hitboxA - hitbox of object A
         * @param	objectA - first object in the collision-check pair
         * @param   hitboxB - hitbox of object B
         * @param	objectB - second object in the collision-check pair
         * @param	deltaTime - length of the time step in which we are checking collisions
         * @return  A CollisionInfo if there was a collision.
         */
        public static function checkHitboxCollision( hitboxA:HitBox,
                                                     objectA:GameObject,
                                                     hitboxB:HitBox,
                                                     objectB:GameObject,
                                                     deltaTime:Number ):CollisionInfo
        {
            var avg_a_dx:Number = objectA.dx + 0.5 * objectA.ddx * deltaTime;
            var avg_b_dx:Number = objectB.dx + 0.5 * objectB.ddx * deltaTime;
            var avg_a_dy:Number = objectA.dy + 0.5 * objectA.ddy * deltaTime;
            var avg_b_dy:Number = objectB.dy + 0.5 * objectB.ddy * deltaTime;
            var aMotionRelativeToB_x:Number = (avg_a_dx - avg_b_dx) * deltaTime;
            var aMotionRelativeToB_y:Number = (avg_a_dy - avg_b_dy) * deltaTime;
            var ci:CollisionInfo = null;

            _tempPoint1.x = objectA.x;
            _tempPoint1.y = objectA.y;
            _tempPoint2.x = objectB.x;
            _tempPoint2.y = objectB.y;

            if (quickCollisionPossibilityCheck( hitboxA,
                                                _tempPoint1,
                                                hitboxB,
                                                _tempPoint2,
                                                aMotionRelativeToB_x,
                                                aMotionRelativeToB_y) )
            {
                ci = CollisionUtil.internalCheckAabbCollision(
                    hitboxA,
                    _tempPoint1,
                    hitboxB,
                    _tempPoint2,
                    aMotionRelativeToB_x,
                    aMotionRelativeToB_y);
            }
            return ci;
        }

        /**
         * check for collisions between two rects
         * @param   a           hitbox for one of the objects
         * @param   a_offset    translation to apply to a before checking for overlap
         * @param   b           hitbox for the other
         * @param   b_offset  translation to apply to a before checking for overlap
         */
        public static function checkRectCollision( a:HitBox,
                                                   a_offset:Point,
                                                   b:HitBox,
                                                   b_offset:Point ):CollisionInfo
        {
            var a2:HitBox = _tempHitBox1;
            a2.makeIntoCopyOf(a);
            a2.offsetPoint(a_offset);
            var b2:HitBox = _tempHitBox2;
            b2.makeIntoCopyOf(b);
            b2.offsetPoint(b_offset);

            var hitInX:Boolean = (b2.left < a2.right) && (a2.left < b2.right),
                hitInY:Boolean = (b2.top < a2.bottom) && (a2.top < b2.bottom);

            // out now if it's no good
            if (!hitInX && !hitInY)
            {
                return null;
            }

            var aCenterX:Number = a2.left + (0.5 * a2.width),
                aCenterY:Number = a2.top + (0.5 * a2.height),
                bCenterX:Number = b2.left + (0.5 * b2.width),
                bCenterY:Number = b2.top + (0.5 * b2.height),
                xOverlap:Number = 0,
                yOverlap:Number = 0;

            if (hitInX)
            {
                if ( (a2.left <= b2.left) && (b2.right <= a2.right) )
                {
                    xOverlap = b2.width;
                }
                else if ( (a2.left >= b2.left) && (b2.right <= a2.right) )
                {
                    xOverlap = Math.min(b2.right - a2.left, a2.right - b2.left);
                }
                else if (a2.left <= b2.left)
                {
                    xOverlap = a2.right - b2.left;
                }
                else
                {
                    xOverlap = a2.right - a2.left;
                }
            }

            if (hitInY)
            {
                if ( (a2.top <= b2.top) && (b2.bottom <= a2.bottom) )
                {
                    yOverlap = b2.width;
                }
                else if ( (a2.top >= b2.top) && (b2.bottom <= a2.bottom) )
                {
                    yOverlap = Math.min(b2.bottom - a2.top, a2.bottom - b2.top);
                }
                else if (a2.top <= b2.top)
                {
                    yOverlap = a2.bottom - b2.top;
                }
                else
                {
                    yOverlap = a2.bottom - a2.top;
                }
            }
            var xCollision:Boolean = hitInY && (xOverlap < a2.width) &&
                                     ( (hitInX && (xOverlap < yOverlap)) );
            var yCollision:Boolean = !xCollision && hitInX && hitInY;

            var ci:CollisionInfo = null;
            if (yCollision)
            {
                ci = new CollisionInfo();
                ci.location.x = aCenterX;
                ci.location.y = (aCenterY > bCenterY) ? b2.bottom : b2.top;
                ci.normal.x = 0;
                ci.normal.y = (aCenterY > bCenterY) ? 1.0 : -1.0;
                ci.hitbox1 = a;
                ci.hitbox2 = b;
            }
            else if (xCollision)
            {
                ci = new CollisionInfo();
                ci.location.x = (aCenterX > bCenterX) ? b2.right : b2.left;
                ci.location.y = aCenterY;
                ci.normal.x = (aCenterX > bCenterX) ? 1.0 : -1.0;
                ci.normal.y = 0;
                ci.hitbox1 = a;
                ci.hitbox2 = b;
            }

            return ci;
        }

        /**
         * check for collisions between two potentially moving rects
         */
        private static function internalCheckAabbCollision(
            objectAHitBox:HitBox,
            objectAOffset:Point,
            objectBHitBox:HitBox,
            objectBOffset:Point,
            aMotionRelativeToB_x:Number,
            aMotionRelativeToB_y:Number ):CollisionInfo
        {
            var ci:CollisionInfo = null;
            var dist:Number = Math.sqrt( (aMotionRelativeToB_x*aMotionRelativeToB_x) + (aMotionRelativeToB_y*aMotionRelativeToB_y) );
            var steps:Number = 1 + Math.floor(dist / COLLISION_STEP_SIZE_MAX);

            for (var step:Number = 0; step <= steps; ++step)
            {
                var scale:Number = (step) / steps;
                _tempPoint3.x = objectAOffset.x + aMotionRelativeToB_x * scale;
                _tempPoint3.y = objectAOffset.y + aMotionRelativeToB_y * scale;

                ci = checkRectCollision( objectAHitBox,
                                         _tempPoint3,
                                         objectBHitBox,
                                         objectBOffset );
                if (ci)
                {
                    ci.t = step / steps;
                    break;
                }
            }
            return ci;
        }

        /**
         * quick check for whether there is any chance of collision
         */
        private static function quickCollisionPossibilityCheck(
            objectABounds:Rectangle,
            objectAOffset:Point,
            objectBBounds:Rectangle,
            objectBOffset:Point,
            aMotionRelativeToB_x:Number,
            aMotionRelativeToB_y:Number ):Boolean
        {
            // too far left
            var aMaxX:Number = Math.max(objectABounds.right, objectABounds.right + aMotionRelativeToB_x);
            if (objectAOffset.x + aMaxX < objectBOffset.x + objectBBounds.left) { return false; }

            // too far right
            var aMinX:Number = Math.min(objectABounds.left, objectABounds.left + aMotionRelativeToB_x);
            if (objectAOffset.x + aMinX > objectBOffset.x + objectBBounds.right) { return false; }

            // too far up
            var aMaxY:Number = Math.max(objectABounds.bottom, objectABounds.bottom + aMotionRelativeToB_y);
            if (objectAOffset.y + aMaxY < objectBOffset.y + objectBBounds.top) { return false; }

            // too far down
            var aMinY:Number = Math.min(objectABounds.top, objectABounds.top + aMotionRelativeToB_y);
            if (objectAOffset.y + aMinY > objectBOffset.y + objectBBounds.bottom) { return false; }

            return true;
        }
    }
}
