package stego.util.graphics
{
    import flash.display.Bitmap;
    import flash.display.BitmapData;
    import flash.display.BlendMode;
    import flash.geom.ColorTransform;
    import flash.geom.Matrix;
    import flash.geom.Point;
    import flash.geom.Rectangle;
    import stego.util.MathUtil;
    import stego.util.ObjectPool;

    /**
     * wraps a Bitmap to more easily draw on it
     * @author Mike Meyer
     */
    public class Canvas
    {
        /** width of the entire tileset image */
        public function get width():int { return _width; }
        public function set width(value:int):void
        {
            if (MathUtil.roundUpToPowerOf2(_width) < value)
            {
                _bufferUpToDate = false;
            }
            _width = value;
        }
        protected var _width:int;

        /** height of the entire tileset image */
        public function get height():int { return _height; }
        public function set height(value:int):void
        {
            if (MathUtil.roundUpToPowerOf2(_height) < value)
            {
                _bufferUpToDate = false;
            }
            _height = value;
        }
        protected var _height:int;

        /** Bitmap object it draws on */
        public function get target():Bitmap { return _target; }
        public function set target(value:Bitmap):void
        {
            if (_target != value)
            {
                _target = value;
                _bufferUpToDate = false;
            }
        }
        protected var _target:Bitmap;

        /** rect for the full canvas */
        public function get rect():Rectangle { return _target.bitmapData.rect; }

        /** access to the BitmapData in case you really want to mess with it directly */
        public function get bitmapData():BitmapData
        {
            resizeBufferIfNeeded();
            return _target.bitmapData;
        }
        public function set bitmapData(value:BitmapData):void
        {
            if (_target && _target.bitmapData != value)
            {
                if (_ownsBitmapData && _target.bitmapData)
                {
                    ObjectPool.returnBitmapData(_target.bitmapData);
                }
                _target.bitmapData = value;
                _width = value.width;
                _height = value.height;
                _bufferUpToDate = true;
                _ownsBitmapData = false;
            }
        }

        /** whether the current buffer needs rebuilding before we can redraw individual tiles */
        protected var _bufferUpToDate:Boolean = false;

        /** whether the canvas has already been cleared */
        protected var _isClear:Boolean = true;

        /**
         * whether the canvas created the bitmap data,
         * and thus whether it should be returned to the pool
         */
        protected var _ownsBitmapData:Boolean = false;

        /**
         * prepare the canvas for use
         */
        public function Canvas(target:Bitmap = null, width:int = 32, height:int = 32):void
        {
            _width = width;
            _height = height;
            _target = target;
        }

        /**
         * clean up, kill any references this has to other things.
         */
        public function destroy():void
        {
            if (_target)
            {
                if (_ownsBitmapData && _target.bitmapData)
                {
                    ObjectPool.returnBitmapData(_target.bitmapData);
                    _target.bitmapData = null;
                }
                _ownsBitmapData = false;
            }
            _target = null;
            _bufferUpToDate = false;
            _isClear = true;
        }

        /** no point in new'ing these all the time */
        protected static var _tempPoint:Point = new Point();
        /** no point in new'ing these all the time */
        protected static var _tempMatrix:Matrix = new Matrix();

        /**
         * clear whatever has been drawn to the canvas
         */
        public function clear():void
        {
            if (!_isClear && _target && _target.bitmapData)
            {
                _target.bitmapData.fillRect(_target.bitmapData.rect, 0);
                _isClear = true;
            }
        }

        /**
         * fades out the bitmap contents
         * @param   amount  how much to fade the canvas by. [0-1] 0 is not at all. 1 makes it completely transparent
         */
        public function fade(amount:Number):void
        {
            if (!_isClear && _target && _target.bitmapData)
            {
                var ct:ColorTransform = ObjectPool.getObject(ColorTransform);
                ct.alphaMultiplier = 1 - amount;
                ct.alphaOffset = 0;
                ct.redMultiplier = 1;
                ct.redOffset = 0;
                ct.greenMultiplier = 1;
                ct.greenOffset = 0;
                ct.blueMultiplier = 1;
                ct.blueOffset = 0;
                _target.bitmapData.colorTransform(_target.bitmapData.rect, ct);
                ObjectPool.returnObject(ColorTransform, ct);
            }
        }

        /**
         * fades out the bitmap contents
         * color values should be between zero and one
         * @param   amount  how much to fade the canvas by. [0-1] 0 is not at all. 1 makes it completely transparent
         */
        public function fadeTowardColor( red:Number,
                                         green:Number,
                                         blue:Number,
                                         alpha:Number,
                                         amount:Number,
                                         rect:Rectangle = null ):void
        {
            if (_target && _target.bitmapData)
            {
                if (null == rect) { rect = _target.bitmapData.rect; }
                var ct:ColorTransform = ObjectPool.getObject(ColorTransform);
                var m:Number = 1.0 - amount;
                ct.alphaMultiplier = m;
                ct.alphaOffset = amount * alpha * 255;
                ct.redMultiplier = m;
                ct.redOffset = amount * red * 255;
                ct.greenMultiplier = m;
                ct.greenOffset = amount * green * 255;
                ct.blueMultiplier = m;
                ct.blueOffset = amount * blue * 255;
                _target.bitmapData.colorTransform(rect, ct);
                ObjectPool.returnObject(ColorTransform, ct);
            }
        }

        /**
         * Fills the rectangle area of the canvas with the texture.
         * @param	bitmapData      bitmapdata to draw.
         * @param	sourceRect      what part of the source bitmap to use
         * @param	destRect        what part of the canvas to draw to
         * @param	clobberAlpha    If set to true, then instead of doing alpha
         *                          blending, the alpha in the source image
         *                          will be copied into the canvas.
         */
        public function copyPixels(bitmapData:BitmapData, sourceRect:Rectangle, destPoint:Point, clobberAlpha:Boolean = false):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.copyPixels(bitmapData, sourceRect, destPoint, null, null, !clobberAlpha);
            _isClear = false;
        }
        /**
         * Fills the rectangle area of the canvas with the texture.
         * @param	bitmapData      bitmapdata to draw.
         * @param	sourceRect      what part of the source bitmap to use
         * @param	matrix          matrix to use to transform the image (or null)
         * @param	colorTransform  colorTransform to apply (or null)
         * @param	blendMode       value from the flash.display.BlendMode class for how to blend the image in (or null)
         * @param   smoothing       whether to smooth the image when transformed.
         */
        public function drawBitmap( bitmapData:BitmapData,
                                    sourceRect:Rectangle = null,
                                    matrix:Matrix = null,
                                    colorTransform:ColorTransform = null,
                                    blendMode:String = null,
                                    smoothing:Boolean = false ):void
        {
            resizeBufferIfNeeded();

            if (!sourceRect) { sourceRect = bitmapData.rect; }

            var buffer:BitmapData = ObjectPool.getBitmapData(sourceRect.width, sourceRect.height);
            _tempPoint.x = 0;
            _tempPoint.y = 0;
            buffer.copyPixels(bitmapData, sourceRect, _tempPoint);
            _target.bitmapData.draw(buffer, matrix, colorTransform, blendMode, null, smoothing);
            ObjectPool.returnBitmapData(buffer);

            _isClear = false;
        }

        /**
         * Fills the canvas with a solid color.
         * @param   color   ARGB value to fill with
         */
        public function fill(color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.fillRect(_target.bitmapData.rect, color);
            _isClear = false;
        }

        /**
         * Fills the canvas with a solid color.
         * @param   color   RGB value to fill with
         */
        public function fillRgb(color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.fillRect(_target.bitmapData.rect, 0xFF000000|color);
            _isClear = false;
        }

        /**
         * Fills the rectangle area of the canvas with a solid color.
         * @param	rect    area to fill
         * @param   color   ARGB value to fill with
         */
        public function fillRect(rect:Rectangle, color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.fillRect(rect, color);
            _isClear = false;
        }

        /**
         * Fills the rectangle area of the canvas with a solid color.
         * @param	rect    area to fill
         * @param   color   RGB value to fill with
         */
        public function fillRectRgb(rect:Rectangle, color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.fillRect(rect, 0xFF000000|color);
            _isClear = false;
        }

        /**
         * Set just one pixel of the canvas
         * @param	x       x position
         * @param   y       y position
         * @param   color   ARGB color to set it to
         */
        public function setPixel(x:int, y:int, color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.setPixel32(x,y,color);
            _isClear = false;
        }

        /**
         * Set just one pixel of the canvas
         * @param	x       x position
         * @param   y       y position
         * @param   color   RGB color to set it to
         */
        public function setPixelRgb(x:int, y:int, color:uint):void
        {
            resizeBufferIfNeeded();
            _target.bitmapData.setPixel32(x,y,0xFF000000|color);
            _isClear = false;
        }

        /**
         * make sure that our buffer is ready to be shown/drawn on
         */
        protected function resizeBufferIfNeeded():void
        {
            if (!_target.bitmapData || !_bufferUpToDate)
            {
                var oldBitmapData:BitmapData = _target.bitmapData;
                _target.bitmapData = ObjectPool.getBitmapData( MathUtil.roundUpToPowerOf2(_width),
                                                               MathUtil.roundUpToPowerOf2(_height) );
                _target.bitmapData.fillRect(_target.bitmapData.rect, 0);
                _ownsBitmapData = true;
                _tempPoint.x = 0;
                _tempPoint.y = 0;
                if (oldBitmapData)
                {
                    _target.bitmapData.copyPixels(oldBitmapData, oldBitmapData.rect, _tempPoint);
                    if (_ownsBitmapData)
                    {
                        ObjectPool.returnBitmapData(oldBitmapData);
                    }
                }
                _bufferUpToDate = true;
            }
        }

        /**
         * draw a solid-color rect onto the canvas with the opacity given by the alpha channel
         */
        public function fillRectBlended(rect:Rectangle, argb_color:uint):void
        {
            if ( (rect.width != 0) && (rect.height != 0) )
            {
                resizeBufferIfNeeded();
                var bd:BitmapData = ObjectPool.getBitmapData(rect.width, rect.height);
                bd.fillRect(bd.rect, argb_color);
                _tempMatrix.identity();
                _tempMatrix.translate(rect.x, rect.y);
                _target.bitmapData.draw(bd, _tempMatrix, null, BlendMode.NORMAL);
                _isClear = false;
                ObjectPool.returnBitmapData(bd);
            }
        }
    }

}
