package stego.util.number
{
    import stego.util.Easing;
    import stego.util.MathUtil;

	/**
     * represent a value that changes over time
     * @author Mike Meyer
     */
    public class Animated
    {
        /** constant to represent how it loops */
        public static const PINGPONG:String = "pingpong";
        public static const REPEAT:String = "repeat";

        /** start value and one extreme of the range it animates in */
        public var start:Number;

        /** one extreme of the range it animates in. "end" value if it isn't looping */
        public var end:Number;

        /** whether to loop the animation */
        public var loop:Boolean;

        /**
         * how exactly it loops to loop the animation.
         * should be one of the static consts in stego.util.number.Animateds
         */
        public var loopStyle:String;

        /** easing function for how the value changes over time */
        public var easing:Function;

        /** get/set how far into the animation this is */
        public function get t():Number { return _t; }
        public function set t(value:Number):void
        {
            _t = value;
        }
        protected var _t:Number;

        /**
         * how long a full loop of the value would take
         * note that for ping pong loop style, this is
         * only the time it takes to go ONE way
         */
        public function get period():Number { return _period; }
        public function set period(value:Number):void { _period = value; update(0); }
        private var _period:Number;


        /**
         * how many times per second the value goes through its animation
         * note that for ping pong loop style, this is counting each direction
         * as an entire loop
         */
        public function get frequency():Number { return 1.0 / _period; }
        public function set frequency(value:Number):void { _period = 1.0 / value; update(0); }

        /**
         * current value of the animation
         */
        public function get value():Number { return _value; }
        protected var _value:Number;

        /**
         * constructor
         *
         * @param	start       start value and one extreme of the range it
         *                      animates in
         * @param	end         one extreme of the range it animates in. "end"
         *                      value if it isn't looping
         * @param	loop        whether to loop the animation
         * @param	loopStyle   how exactly it loops to loop the animation.
         *                      use one of the static consts in
         *                      stego.util.number.Animated
         * @param	easing      easing function for how the value changes over
         *                      time. defaults to stego.util.Easing.quadInOut
         * @param	period      how long a full loop of the value would take
         *                      note that for ping pong loop style, this is
         *                      only the time it takes to go ONE way
         * @param   startTime   how many seconds into the animation it should
         *                      start
         */
        public function Animated( start:Number = 0,
                                  end:Number = 1,
                                  loop:Boolean = true,
                                  loopStyle:String = "pingpong",
                                  easing:Function = null,
                                  period:Number = 1,
                                  startTime:Number = 0 )
        {
            this.start = start;
            this.end = end;
            this.loop = loop;
            this.loopStyle = loopStyle;
            this.easing = (null == easing) ? Easing.quadInOut : easing;
            this._period = period;
            _t = startTime;
            _value = update(0);
        }

        /**
         * animate and return the new value
         */
        public function update(deltaTime:Number):Number
        {
            _t = MathUtil.wrap(_t + deltaTime, 0, 2.0 * _period);
            switch (loopStyle)
            {
                case REPEAT:
                    var lerpVal:Number = easing( MathUtil.wrap(_t, 0, _period) / _period );
                    _value = MathUtil.lerp(lerpVal, start, end);
                    break;
                case PINGPONG:
                default:
                    _value = MathUtil.remap( MathUtil.pingpong(_t, 0, _period, easing),
                                             0,
                                             _period,
                                             start,
                                             end );
                    break;
            }
            return _value;
        }
    }

}
